#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <getopt.h>
#include <stdint.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <errno.h>
#include "EcatDeviceAPI.h"

#define INPUT_ARG_SIZE 1024
#define LINE_LOOP_POS1 10
#define LINE_LOOP_POS2 -10
#define LINE_LOOP_VEL 10

#define TASK_IDLE_INDEX 1000

char input_arg[INPUT_ARG_SIZE+1];
uint16_t device_no = 0;
uint32_t link_up, slaves_responding, al_states, wc;
uint16_t axis_cnt;
uint16_t servo_motor_slave_no = 0;
uint16_t mc_task_idx = 0;
uint16_t mc_sub_task_idx = 0;
uint8_t mc_sub_task_ok = 0;
uint16_t mc_reset_task_idx = 1000;
uint16_t mc_task_time_out;
uint8_t mc_task_error;
uint8_t should_stop = 0;

char mc_error_msg[256] = "";
uint16_t run_mode = 0;

/*****************************************************************************/
static void sig_handler(int sig)
{
	should_stop = 1;
}
/*****************************************************************************/
void show_device_state(uint32_t *link_up, uint32_t *slaves_responding, uint32_t *al_states, uint32_t *wc)
{
	char StrAlState[255];

	if(*al_states == 0)
		sprintf(StrAlState,"INIT");
	else if(*al_states == 0x02)
		sprintf(StrAlState,"PREOP");
	else if(*al_states == 0x04)
		sprintf(StrAlState,"SAFEOP");
	else if(*al_states == 0x08)
		sprintf(StrAlState,"OP");
	else
		sprintf(StrAlState,"Invalid");


	printf("Device State\n");
	printf("\t |_  Slave(s): %u\n", *slaves_responding);
	printf("\t |_ AL States: %s\n", StrAlState);
	printf("\t |_      Link: %s\n", *link_up? "up" : "down");
	printf("\t |_        Wc: %-u\n", *wc);
	printf("\n");
}
/*****************************************************************************/
void show_act_list(void)
{
	printf("/**************action**************/\n");
	printf("[1]  - Axes Error Reset\n"
			"[2]  - Restart MC Task\n"
			"[3]  - Home Task\n"
			"[4]  - Stop Task\n"
			"[5]  - Line Task\n"
			"[6]  - 2D Circular Task\n"
			"[7]  - 3D Circular Task\n"
			"[8]  - 2D Helical Task\n"
			"[9]  - 3D Helical Task\n"
			);
}
/*****************************************************************************/
void show_mc_task_info()
{
	printf("Motion Control Task\n");
	switch(run_mode) {
	case 0:
		printf("\t |_   Task: Line\n");
		break;
	case 1:
		printf("\t |_   Task: 2D Circular\n");
		break;
	case 2:
		printf("\t |_   Task: 3D Circular\n");
		break;
	case 3:
		printf("\t |_   Task: 2D Helical\n");
		break;
	case 4:
		printf("\t |_   Task: 3D Helical\n");
		break;
	case 100:
		printf("\t |_   Task: Home\n");
		break;
	case 101:
		printf("\t |_   Task: Stopping\n");
		break;
	case 1000:
		printf("\t |_   Task: IDLE\n");
		break;
	default:
		printf("\t |_   Task: Invalid\n");
		break;
	}
	if (mc_task_error)
		printf("\t |_  Index: error,%u\n", mc_reset_task_idx);
	else
		printf("\t |_  Index: main,%u\n", mc_task_idx);
	printf("\t |_  Error: %s\n", mc_error_msg);
}
/*****************************************************************************/
void handle_act(void)
{
	uint32_t act_no = atoi(input_arg);
	switch(act_no)
	{
	case 1:
		mc_reset_task_idx = 0;
		mc_task_error = 0;
		mc_sub_task_idx = 0;
		mc_sub_task_ok = 0;
		break;
	case 2:
		memset(mc_error_msg, 0, sizeof(mc_error_msg));
		mc_task_idx = 0;
		mc_task_error = 0;
		break;
	case 3:
	case 4:
	case 5:
	case 6:
	case 7:
	case 8:
	case 9:
		memset(mc_error_msg, 0, sizeof(mc_error_msg));
		mc_task_idx = 0;
		mc_task_error = 0;
		if (act_no == 3)
			run_mode = 100;
		else if (act_no == 4)
			run_mode = 101;
		else if (act_no == 5)
			run_mode = 0;
		else if (act_no == 6)
			run_mode = 1;
		else if (act_no == 7)
			run_mode = 2;
		else if (act_no == 8)
			run_mode = 3;
		else if (act_no == 9)
			run_mode = 4;
		break;
	default:
		printf("Invalid action number\n");
		break;
	}


}
/*****************************************************************************/
int __sleep(void)
{
	struct timeval tv;
	fd_set readfds;
	int syscall_ret;
	ssize_t read_ret;
	tv.tv_sec = 0;
	tv.tv_usec = 1000;
	FD_ZERO(&readfds);
	FD_SET(STDIN_FILENO, &readfds);
	syscall_ret = select(STDIN_FILENO+1, &readfds, NULL, NULL, &tv);
	if(!syscall_ret) {

		return syscall_ret;

	} else {
		if(FD_ISSET(STDIN_FILENO,&readfds)) {
			if (errno != EINTR) {
				read_ret = read(STDIN_FILENO, input_arg, INPUT_ARG_SIZE);
				if(read_ret) {
					show_act_list();
					if (fgets(input_arg, INPUT_ARG_SIZE , stdin) != NULL)
						handle_act();
				}
			}
		}
	}
	return syscall_ret;
}
/*****************************************************************************/
int32_t check_axis_standstill(uint16_t axis_no)
{
	int32_t ret;
	uint32_t u32_value;
	ret = ECAT_McGetAxisState_Ex(device_no, axis_no, &u32_value);
	if (!ret) {
		if (u32_value == MC_AS_STANDSTILL)
			return 0;
		else if (u32_value == MC_AS_ERRORSTOP)
			return -1;
		else
			return 1;
	} else {
		sprintf(mc_error_msg, "Failed to get axis%u state:%d", axis_no, ret);
		mc_task_error = 1;
		return -1;
	}
}
/*****************************************************************************/
int32_t check_group_standby()
{
	int32_t ret;
	uint32_t u32_value;
	ret = ECAT_McGetGroupState(device_no, 0, &u32_value);
	if (!ret) {
		if (u32_value == MC_GS_STANDBY)
			return 0;
		else if (u32_value == MC_GS_ERRORSTOP)
			return -1;
		else
			return 1;
	} else {
		sprintf(mc_error_msg, "Failed to get group0 state:%d", ret);
		mc_task_error = 1;
		return -1;
	}
}
/*****************************************************************************/
void check_group_state(uint16_t task_idx)
{
	int32_t ret;
	ret = check_group_standby();
	if (ret == 0)
		mc_task_idx = task_idx;
	else if (ret == -1) {
		if (mc_task_error)
			return;

		sprintf(mc_error_msg, "Invalid axis state");
		mc_task_error = 1;
	}
}
/*****************************************************************************/
void mc_task_error_reset(void)
{
	int32_t ret;
	uint16_t i;
	int8_t to_next = 1;

	switch(mc_reset_task_idx) {
		case 0:
			for (i = 0; i < axis_cnt; i++ ) {
//				ret = ECAT_McAxisErrorReset(device_no, i);
				//In some cases, the motor driver will servo off when it encounters an error
				//user can use ECAT_McSetAxisServoOn instead of ECAT_McAxisErrorReset
				//ECAT_McSetAxisServoOn also has the function of clearing errors
				ret = ECAT_McSetAxisServoOn(device_no, i, 1);
				if (ret) {
					sprintf(mc_error_msg, "Failed to set axis%u Servo ON:%d", i, ret);
					mc_reset_task_idx = 1000;
					return;
				}
			}
			mc_reset_task_idx++;
			mc_task_time_out = 500;
			break;
		case 1:
			for (i = 0; i < axis_cnt; i++ ) {
				ret = check_axis_standstill(i);
				if (ret == -1) {
					if (mc_task_error)
						return;

					to_next = 0;
				}
			}

			if (to_next) {
				memset(mc_error_msg, 0, sizeof(mc_error_msg));
				mc_reset_task_idx = 1000;
				mc_task_error = 0;
			} else {
				if (mc_task_time_out)
					mc_task_time_out--;
				else {
					sprintf(mc_error_msg, "Axis error reset timeout(axis state error)");
					mc_reset_task_idx = 1000;
				}
			}
			break;
		default:
			break;
	}
}
/*****************************************************************************/
void mc_sub_task_servo_on(void)
{
	int32_t ret;
	uint32_t u32_value;
	uint16_t i;
	int8_t to_next = 1;

	switch(mc_sub_task_idx) {
		case 0:
			for (i = 0; i < axis_cnt; i++ ) {
				ret = ECAT_McSetAxisServoOn(device_no, i, 1);
				if (ret) {
					sprintf(mc_error_msg, "Failed to set axis%u Servo ON:%d", i, ret);
					mc_task_error = 1;
					return;
				}
			}
			mc_sub_task_idx++;
			mc_task_time_out = 500;
			break;
		case 1:
			for (i = 0; i < axis_cnt; i++ ) {
				ret = check_axis_standstill(i);
				if (ret == -1) {
					if (mc_task_error)
						return;
					to_next = 0;
				}
			}

			if (to_next) {
				mc_sub_task_idx++;
				mc_task_time_out = 500;
			}
			else {
				if (mc_task_time_out)
					mc_task_time_out--;
				else {
					sprintf(mc_error_msg, "Invalid axis state");
					mc_task_error = 1;
				}
			}
			break;
		case 2:
			for (i = 0; i < axis_cnt; i++ ) {
				ret = ECAT_McGetAxisDI_Ex(device_no, i, &u32_value);
				if (!ret) {
					if (!(u32_value & 0x20)) {
						if (mc_task_time_out) {
							mc_task_time_out--;
							to_next = 0;
						} else {
							strcpy(mc_error_msg, "Axis DI error(SVN)");
							mc_task_error = 1;
							return;
						}
					}
				} else {
					sprintf(mc_error_msg, "Failed to get axis%u DI:%d", i, ret);
					mc_task_error = 1;
					return;
				}
			}

			if (to_next) {
				mc_sub_task_idx++;
				mc_sub_task_ok = 1;
			}
			break;

		default:
			break;
	}
}
/*****************************************************************************/
void mc_task_home(void)
{
	int32_t ret;
	uint16_t i;
	int8_t to_next = 1;

	switch(mc_task_idx) {
	case 0:
		mc_sub_task_idx = 0;
		mc_sub_task_ok = 0;
		mc_task_idx++;
		break;
	case 1:
		if (mc_sub_task_ok)
			mc_task_idx = 3;
		else
			mc_sub_task_servo_on();
		break;
	case 3:
		check_group_state(mc_task_idx + 1);
		break;
	case 4:
		for (i = 0; i < axis_cnt; i++ ) {
			ret = ECAT_McAxisHome(device_no, i);
			if (ret) {
				sprintf(mc_error_msg, "Axis%u move home failed:%d", i, ret);
				mc_task_error = 1;
				return;
			}
		}
		mc_task_idx++;
		break;
	case 5:
		for (i = 0; i < axis_cnt; i++ ) {
			ret = check_axis_standstill(i);
			if (ret == -1) {
				if (mc_task_error)
					return;

				to_next = 0;
			}
		}

		if (to_next) {
			mc_task_idx++;
			run_mode = TASK_IDLE_INDEX;
		}
		break;
	default:
		break;
	}
}
/*****************************************************************************/
void mc_task_group_stop(void)
{
	int32_t ret;

	switch(mc_task_idx) {
		case 0:
			ret = ECAT_McGroupStop(device_no, 0);
			if (ret) {
				sprintf(mc_error_msg, "Group0 stop failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 1:
			check_group_state(mc_task_idx + 1);
			break;
		case 2:
			mc_task_idx++;
			run_mode = TASK_IDLE_INDEX;
			break;
		default:
			break;
	}

}
/*****************************************************************************/
void mc_task_line(void)
{
	int32_t ret;
	double line_pos[MC_AXIS_NO_MAX] = {0};

	switch(mc_task_idx) {
		case 0:
			mc_sub_task_idx = 0;
			mc_sub_task_ok = 0;
			mc_task_idx++;
			break;
		case 1:
			if (mc_sub_task_ok)
				mc_task_idx = 3;
			else
				mc_sub_task_servo_on();
			break;
		case 3:
			check_group_state(mc_task_idx + 1);
			break;
		case 4:
			line_pos[0] = LINE_LOOP_POS1;
			line_pos[1] = LINE_LOOP_POS1;
			line_pos[2] = LINE_LOOP_POS1;
			ret = ECAT_McGroupMoveLineRel_Ex(device_no, 0, line_pos, LINE_LOOP_VEL);
			if (ret) {
				sprintf(mc_error_msg, "Group0 move line relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 5:
			check_group_state(mc_task_idx + 1);
			mc_task_time_out = 2000;
			break;
		case 6:
			if (mc_task_time_out)
				mc_task_time_out--;
			else
				mc_task_idx++;
			break;
		case 7:
			line_pos[0] = LINE_LOOP_POS2;
			line_pos[1] = LINE_LOOP_POS2;
			line_pos[2] = LINE_LOOP_POS2;
			ret = ECAT_McGroupMoveLineRel_Ex(device_no, 0, line_pos, LINE_LOOP_VEL);
			if (ret) {
				sprintf(mc_error_msg, "Group0 move line relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 8:
			check_group_state(mc_task_idx + 1);
			mc_task_time_out = 2000;
			break;
		case 9:
			if (mc_task_time_out)
				mc_task_time_out--;
			else
				mc_task_idx = 4;
			break;
		default:
			break;
	}

}
/*****************************************************************************/
void mc_task_2d_circ(void)
{
	int32_t ret;
	double pos[MC_AXIS_NO_MAX] = {0};

	switch(mc_task_idx) {
		case 0:
			mc_sub_task_idx = 0;
			mc_sub_task_ok = 0;
			mc_task_idx++;
			break;
		case 1:
			if (mc_sub_task_ok)
				mc_task_idx = 3;
			else
				mc_sub_task_servo_on();
			break;
		case 3:
			check_group_state(mc_task_idx + 1);
			break;
		case 4:
			pos[0] = 20.0;
			pos[1] = 0.0;
			pos[2] = 0.0;
			ret = ECAT_McGroupMoveCircularRel_CP_Angle_Ex(device_no, 0, 10, 225, pos);
			if (ret) {
				sprintf(mc_error_msg, "Group0 move circular relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 5:
			check_group_state(mc_task_idx + 1);
			mc_task_time_out = 2000;
			break;
		case 6:
			if (mc_task_time_out)
				mc_task_time_out--;
			else
				mc_task_idx++;
			break;
		case 7:
			pos[0] = 0.0;
			pos[1] = 30.0;
			pos[2] = 0.0;
			ret = ECAT_McGroupMoveCircularRel_CP_Angle_Ex(device_no, 0, 10, -585, pos);
			if (ret) {
				sprintf(mc_error_msg, "Group0 move circular relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 8:
			check_group_state(mc_task_idx + 1);
			mc_task_time_out = 2000;
			break;
		case 9:
			if (mc_task_time_out)
				mc_task_time_out--;
			else
				mc_task_idx = 4;
			break;
		default:
			break;
	}
}
/*****************************************************************************/
void mc_task_3d_circ(void)
{
	int32_t ret;
	double cp[3] = {0};
	double ep[3] = {0};

	switch(mc_task_idx) {
		case 0:
			mc_sub_task_idx = 0;
			mc_sub_task_ok = 0;
			mc_task_idx++;
			break;
		case 1:
			if (mc_sub_task_ok)
				mc_task_idx = 3;
			else
				mc_sub_task_servo_on();
			break;
		case 3:
			check_group_state(mc_task_idx + 1);
			break;
		case 4:
			cp[0] = 21.213;
			cp[1] = 21.213;
			cp[2] = 0.0;
			ep[0] = 36.21286;
			ep[1] = 36.21286;
			ep[2] = -21.21280;
			ret = ECAT_McGroupMove3DCircularRel_CP_EP_Ex(device_no, 0, 10, 0, cp, ep);
			if (ret) {
				sprintf(mc_error_msg, "Group0 move circular relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 5:
			check_group_state(mc_task_idx + 1);
			mc_task_time_out = 2000;
			break;
		case 6:
			if (mc_task_time_out)
				mc_task_time_out--;
			else
				mc_task_idx++;
			break;
		case 7:
			cp[0] = 21.213;
			cp[1] = 21.213;
			cp[2] = 0.0;
			ep[0] = 1.98749;
			ep[1] = 1.98749;
			ep[2] = -12.67830;
			ret = ECAT_McGroupMove3DCircularRel_CP_EP_Ex(device_no, 0, 10, 0, cp, ep);
			if (ret) {
				sprintf(mc_error_msg, "Group0 move circular relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 8:
			check_group_state(mc_task_idx + 1);
			mc_task_time_out = 2000;
			break;
		case 9:
			if (mc_task_time_out)
				mc_task_time_out--;
			else
				mc_task_idx = 4;
			break;
		default:
			break;
	}
}
/*****************************************************************************/
void mc_task_2d_helical(void)
{
	int32_t ret;
	double pos[MC_AXIS_NO_MAX] = {0};

	switch(mc_task_idx) {
		case 0:
			mc_sub_task_idx = 0;
			mc_sub_task_ok = 0;
			mc_task_idx++;
			break;
		case 1:
			if (mc_sub_task_ok)
				mc_task_idx = 3;
			else
				mc_sub_task_servo_on();
			break;
		case 3:
			check_group_state(mc_task_idx + 1);
			break;
		case 4:
			pos[0] = 20.0;
			pos[1] = 0.0;
			pos[2] = 0.0;
			ret = ECAT_McGroupMoveHelicalRel_Ex(device_no, 0, 720, pos, 5, 10);
			if (ret) {
				sprintf(mc_error_msg, "Group0 move helical relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 5:
			check_group_state(mc_task_idx + 1);
			mc_task_time_out = 2000;
			break;
		case 6:
			if (mc_task_time_out)
				mc_task_time_out--;
			else
				mc_task_idx++;
			break;
		case 7:
			pos[0] = 0.0;
			pos[1] = 30.0;
			pos[2] = 0.0;
			ret = ECAT_McGroupMoveHelicalRel_Ex(device_no, 0, 1080, pos, 5, 10);
			if (ret) {
				sprintf(mc_error_msg, "Group0 move helical relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 8:
			check_group_state(mc_task_idx + 1);
			mc_task_time_out = 2000;
			break;
		case 9:
			if (mc_task_time_out)
				mc_task_time_out--;
			else
				mc_task_idx = 4;
			break;
		default:
			break;
	}
}
/*****************************************************************************/
void mc_task_3d_helical(void)
{
	int32_t ret;
	double cp[3] = {0};
	double nv[3] = {0};

	switch(mc_task_idx) {
		case 0:
			mc_sub_task_idx = 0;
			mc_sub_task_ok = 0;
			mc_task_idx++;
			break;
		case 1:
			if (mc_sub_task_ok)
				mc_task_idx = 3;
			else
				mc_sub_task_servo_on();
			break;
		case 3:
			check_group_state(mc_task_idx + 1);
			break;
		case 4:
			cp[0] = 14.142;
			cp[1] = 14.142;
			cp[2] = 0.0;
			nv[0] = -0.7071;
			nv[1] = 0.7071;
			nv[2] = 0.0;
			ret = ECAT_McGroupMove3DHelicalAbs_CP_Angle_Ex(device_no, 0, 15, 720, 5, cp, nv);
			if (ret) {
				sprintf(mc_error_msg, "Group0 move circular relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 5:
			check_group_state(mc_task_idx + 1);
			mc_task_time_out = 2000;
			break;
		case 6:
			if (mc_task_time_out)
				mc_task_time_out--;
			else
				mc_task_idx++;
			break;
		case 7:
			cp[0] = 14.142;
			cp[1] = 14.142;
			cp[2] = 0.0;
			nv[0] = -0.7071;
			nv[1] = 0.7071;
			nv[2] = 0.0;
			ret = ECAT_McGroupMove3DHelicalAbs_CP_Angle_Ex(device_no, 0, 15, 1080, 5, cp, nv);
			if (ret) {
				sprintf(mc_error_msg, "Group0 move circular relative failed:%d", ret);
				mc_task_error = 1;
			} else {
				mc_task_idx++;
			}
			break;
		case 8:
			check_group_state(mc_task_idx + 1);
			mc_task_time_out = 2000;
			break;
		case 9:
			if (mc_task_time_out)
				mc_task_time_out--;
			else
				mc_task_idx = 4;
			break;
		default:
			break;
	}
}
/*****************************************************************************/
void mc_task(void)
{
	switch(run_mode) {
		case 0:
			mc_task_line();
			break;
		case 1:
			mc_task_2d_circ();
			break;
		case 2:
			mc_task_3d_circ();
			break;
		case 3:
			mc_task_2d_helical();
			break;
		case 4:
			mc_task_3d_helical();
			break;
		case 100:
			mc_task_home();
			break;
		case 101:
			mc_task_group_stop();
			break;
		default:
			break;
	}
}
/*****************************************************************************/
int main(void) {

	int32_t ret;
	int32_t info_time_tick = 0;
	double double_value;
	uint32_t u32_value;
	int32_t i32_value;
	uint16_t u16_value;
	char axis_state[128];
	char grp_state[128];
	uint16_t slave_no[MC_AXIS_NO_MAX];
	uint16_t sub_axis_no[MC_AXIS_NO_MAX];

	uint16_t i, j;

	if (signal(SIGINT, sig_handler) == SIG_ERR){
		printf("Failed to signal handler\n");
		return EXIT_FAILURE;
	}


	ret = ECAT_OpenDevice(device_no);
	if(ret != 0) {
		printf("Failed to open device:%d\n", ret);
		return EXIT_FAILURE;
	}

	/*****This section can be removed if user don't want the output to be turned off*****/
	/*****output will be turned off when state is "not" in OP (DO/AO will be turn off and servo motor will be servo off)*****/
	ECAT_StopDeviceOpTask(device_no);
	for (int i = 0; i < 10; i++)
	{
		ret = ECAT_GetDeviceStateEx(device_no, &link_up, &slaves_responding, &al_states, &wc);
		if (ret == 0 && al_states == ECAT_AS_PREOP)
			break;
		usleep(1000000);// 1 second
	}
	if (ret != 0 || al_states != ECAT_AS_PREOP)
	{
		printf("ECAT_StopDeviceOpTask failed\n");
		goto out_error;
	}
	/***************************************************************************************/

	ret = ECAT_SetSlaveNoType(device_no, SLAVE_NO_TYPE_ALIAS_ID);
	if (ret != 0)
	{
		printf("Set SlaveNo Type failed, ret:%d", ret);
		goto out_error;
	}

	for (int i = 0; i < 10; i++)
	{
		ret = ECAT_StartDeviceOpTask(device_no, 0, DEV_OP_CYCLE_TIME_1MS, 3);
		if (ret == 0)
			break;
		usleep(1000000);
	}
	if(ret < 0)
	{
		if (ret == -1009)
			printf("Op Failed, please use Utility to build Network Information");
		else
			printf("Op Failed: %d", ret);
		goto out_error;
	}

	//should be OP in 30 second
	for (int i = 0; i < 30; i++)
	{
		ret = ECAT_GetDeviceStateEx(device_no, &link_up, &slaves_responding, &al_states, &wc);
		if (ret == 0 && al_states == ECAT_AS_OP)
			break;
		printf("\033[2J"); //clear
		printf("\033[0;0H");
		show_device_state(&link_up, &slaves_responding, &al_states, &wc);
		usleep(1000000);
	}
	if (ret != 0 || al_states != ECAT_AS_OP)
	{
		printf("Op Failed");
		goto out_error;
	}

	/*
	axis_cnt			//How many Axes in the Ethercat network
	slave_no[]			//SlaveNo(Position or Alias) of the Axis module in the ethercat network	//65535 for Virtual axis
	sub_axis_no[]		//Sub Position of the Axis module in the ethercat network //For multiple Axes modules ,exp. Ecat-2094s 0~3

	Please use ECAT_Utility to create Motion Control Parameter File(.motcfg)
	*/
	ret = ECAT_McInit_Ex(device_no, "Auto.motcfg", slave_no, sub_axis_no, &axis_cnt);
	if(ret < 0)
	{
		if(ret == -1027)
		{
			printf("Auto.motcfg Not exist\n");
			printf("Please use ECAT_Utility.exe to create Motion Control Parameter File(.motcfg)");
		}
		else
			printf("Failed to initialize axes:%d", ret);
		goto out_error;
	}

	for (i = 0; i < axis_cnt; i++ ) {
		ret = ECAT_McAddAxisToGroup_Ex(device_no, 0, i);
		if(ret != 0) {
			printf("Failed to add axis%u to group:%d\n", i, ret);
			goto out_error;
		}
	}

	//These parameters have been set in ECAT_McInit_Ex
	/*
	for (i = 0; i < axis_cnt; i++ ) {
		ret = ECAT_McSetAxisPPU(device_no, i, 1);
		if(ret != 0) {
			printf("Failed to set axis[%d] PPU:%d\n", i, ret);
		}

		ret = ECAT_McSetAxisHomeMethod(device_no, i, 37);
		if(ret != 0) {
			printf("Failed to set axis[%d] home method:%d\n", i, ret);
		}

		ret = ECAT_McSetAxisHomeSpeed(device_no, i, 10000, 1000);
		if(ret != 0) {
			printf("Failed to set axis[%d] home speed:%d\n", i, ret);
		}

		ret = ECAT_McSetAxisHomeAcc(device_no, i, 100000);
		if(ret != 0) {
			printf("Failed to set axis[%d] home acc:%d\n", i, ret);
		}
	}
	*/
	for (i = 0; i < axis_cnt; i++ ) {
		ret = ECAT_McSetAxisAccDecUnit(device_no, i, MC_ACC_DEC_MODE_TIME);
		if (ret != 0)
		{
			printf("Failed to set axis[%d] acc unit:%d\n", i, ret);
		}

		ret = ECAT_McSetAxisAccTime(device_no, i, 100);
		if (ret < 0)
		{
			printf("Failed to set axis[%d] acctime:%d\n", i, ret);
		}
		/*
		*the other method to set accdec
		ret = ECAT_McSetAxisAccDecUnit(device_no, AxisNo, MC_ACC_DEC_MODE_RATE);
		if (ret < 0)
		{
			printf("Failed to set axis[%d] acc unit:%d\n", AxisNo, ret);
		}

		ret = ECAT_McSetAxisAccDecRate(device_no, AxisNo, 100000);//user unit/s^2
		if (ret < 0)
		{
			printf("Failed to set axis[%d] acct rate:%d\n", AxisNo, ret);
		}
		*/
	}

	for (i = 0; i < axis_cnt; i++ ) {
		do
		{
			/*
			when error occurred on master or motor driver
			it will take few ms to process error
			in the meantime, ServoOn or ErrorReset will return -1137

			P.S. ServoOn also has the function of clearing errors
			*/
			ret = ECAT_McSetAxisServoOn(device_no, i, 1);
		} while (ret == -1137);
	}

	uint32_t state[MC_AXIS_NO_MAX] = {0};
	for (i = 0; i < axis_cnt; i++ ) {
		for (size_t j = 0; j < 10; j++)
		{
			/*
			it will take few ms to process ServoOn
			when motor driver is ready, axis state should be standstill
			*/
			ret = ECAT_McGetAxisState_Ex(device_no, i, &state[i]);
			if (ret == 0 && state[i] == MC_AS_STANDSTILL)
				break;
			usleep(100000); //100ms
		}
	}
	for (i = 0; i < axis_cnt; i++ ) {
		if (state[i] != MC_AS_STANDSTILL)
		{
			printf("AxisNo[%d]:Failed to Servo On:ret:%dState:%u\n", i, ret, state[i]);
			goto out_error;
		}
	}

	while(!should_stop) {
		if (__sleep() == 0) {

			ret = ECAT_GetDeviceStateEx(device_no, &link_up, &slaves_responding, &al_states, &wc);

			if ((!ret)) {

				if (al_states == 0x08) {
					if (mc_task_error)
						mc_task_error_reset();
					else
						mc_task();
				}

				if (info_time_tick)
					info_time_tick--;
				else {
					printf("\033[2J"); //clear
					printf("\033[0;0H");

					show_device_state(&link_up, &slaves_responding, &al_states, &wc);

					for (i = 0; i < axis_cnt; i++ ) {
						printf("Axis[%u]-Servo\n", i);

						ret = ECAT_McGetAxisActualPos_Ex(device_no, i, &double_value);
						if (!ret)
							printf("\t |_ Position: %f\n", double_value);
						else
							printf("\t |_ Position: ??\n");

						ret = ECAT_McGetAxisActualVel_Ex(device_no, i, &double_value);
						if (!ret)
							printf("\t |_ Velocity: %f\n", double_value);
						else
							printf("\t |_ Velocity: ??\n");

						ret = ECAT_McGetAxisState_Ex(device_no, i, &u32_value);
						if (!ret) {
							switch(u32_value) {
									case MC_AS_DISABLED:
										strcpy(axis_state, "Disabled");
										break;
									case MC_AS_STANDSTILL:
										strcpy(axis_state, "StandStill");
										break;
									case MC_AS_ERRORSTOP:
										strcpy(axis_state, "ErrorStop");
										break;
									case MC_AS_STOPPING:
										strcpy(axis_state, "Stopping");
										break;
									case MC_AS_HOMING:
										strcpy(axis_state, "Homing");
										break;
									case MC_AS_DISCRETEMOTION:
										strcpy(axis_state, "DiscMotion");
										break;
									case MC_AS_CONTINUOUSMOTION:
										strcpy(axis_state, "ContMotion");
										break;
									case MC_AS_SYNCHRONIZEDMOTION:
										strcpy(axis_state, "SyncMotion");
										break;
									default:
										strcpy(axis_state, "Invalid");
										break;
								}
							printf("\t |_    State: %s\n", axis_state);
						} else
							printf("\t |_    State: ??\n");

						ret =  ECAT_McGetAxisLastError_Ex(device_no, i, &i32_value);
						if (!ret)
							printf("\t |_    Error: %d,", i32_value);
						else
							printf("\t |_    Error: ??,");

						ret = ECAT_McGetAxisDriveError_Ex(device_no, i, &u16_value);
						if (!ret)
							printf("%u\n", u16_value);
						else
							printf("??\n");

						ret = ECAT_McGetAxisDI_Ex(device_no, i, &u32_value);
						if (!ret) {
							printf("\t |_       DI: ");
							for (j = 0; j < 7; j++) {
								if (u32_value & (0x01 << j))
									printf("1 ");
								else
									printf("0 ");
							}
							printf("\n");
						}
						else
							printf("\t |_       DI: ??\n");
						printf("\n");
					}

					show_mc_task_info();

					ret = ECAT_McGetGroupState(device_no, 0, &u32_value);
					if (!ret) {
						switch(u32_value) {
								case MC_GS_DISABLED:
									strcpy(grp_state, "Disabled");
									break;
								case MC_GS_STANDBY:
									strcpy(grp_state, "Standby");
									break;
								case MC_GS_ERRORSTOP:
									strcpy(grp_state, "ErrorStop");
									break;
								case MC_GS_STOPPING:
									strcpy(grp_state, "Stopping");
									break;
								case MC_GS_HOMING:
									strcpy(grp_state, "Homing");
									break;
								case MC_GS_MOVING:
									strcpy(grp_state, "Moving");
									break;
								default:
									strcpy(grp_state, "Invalid");
									break;
							}
						printf("\t |_ Group0: %s,", grp_state);
					} else
						printf("\t |_ Group0: ??,");

					ret = ECAT_McGetGroupCmdBuffer(device_no, 0, &u16_value);
					if(!ret)
						printf("%u\n", u16_value);
					else
						printf("??\n");


					info_time_tick = 500;
				}
			}
		}
	}

	ECAT_McAxisQuickStop(device_no, 0);
	usleep(100000);

out_error:
	ret = ECAT_CloseDevice(device_no);
	if(ret != 0) {
		printf("Failed to close device:%d\n", ret);
		return EXIT_FAILURE;
	} else
		printf("Close device\n");
	return EXIT_SUCCESS;
}
